/*----------------------------------------------
   MegaStorm Tile-based Engine - Roberto Prieto
   Copyright (C) 2007-2012 MegaStorm Systems
  ----------------------------------------------
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    Roberto Prieto
    contact@megastormsystems.com
    http://www.megastormsystems.com

    Used software licenses, including the one for this engine, are located
    at /licenses directory. Please, include them on your distributions.

 --------------------------------------------------------------------------
             MSTE Main engine class
    
    Changelog: 
 v1.01 Sept/2012	- Check History.txt
 v1.00Beta Jul/2012 - Check History.txt
 v0.99 June/2012	- Check History.txt
 v0.98 June/2012	- Check History.txt
 v0.95 May/2012		- Check History.txt
 v0.90 July/2007	- Added an easy, flexible and powerful way to use your own custom map
                    - Fixed some bugs.
                    - First public version including two examples and documentation.
 v0.81 July/2007	- Solved some problems on SaveLevel() and LoadLevel().
                    - Fixed and variable tile size.
                    - Full working engine.
                    - Using new CRM32Pro v4.98.
 v0.80 June/2007	- Save and Load native maps to XML or DPF/XML.
 v0.70 May/2007		- Import maps from image files(BMP/PNG).
                    - Export maps to image files(BMP/PNG).
 v0.60 April/2007	- Native support up to 5 layers. Easy update to n layers.
                    - Full and customizable parallax scroll support.
                    - Independent or group fixed, with scroll or with parallax scroll layers.
                    - Viewport and target surface can be selected in realtime.
                    - OnDrawEnd function event for each layer.
                    - MapData inherit to allow you the creation of custom layers.
   
 -------------------------------------------------------------------------- */
#ifndef _MSTE_H_
#define _MSTE_H_
#define MSTE_VERSION "1.10"

//-------------------------------------------------------------------------------
// Includes.
//-------------------------------------------------------------------------------
#include <CRM32Pro.h>
class cMSTE_MapBase;

// -----------------------------------------------------------------------------

/** @defgroup MSTE MS Tile-based Engine v1.01 - Engine interface and defines 
\image html logotipo_mste.png
* <br>This is the main interface of the scroll engine.
* - Created on 31 July 2007
* - Last modified on 6 September 2012*/
// -----------------------------------------------------------------------------
//! @{  

//-------------------------------------------------------------------------------
// Defines.
//-------------------------------------------------------------------------------
#define TE_LAYERFLAG_DISABLE		 0 //!< Disable all layer's flags. Value by default.
#define TE_LAYERFLAG_LOOPX			 1 //!< Enable the loop on X axis
#define TE_LAYERFLAG_LOOPY			 2 //!< Enable the loop on Y axis
#define TE_LAYERFLAG_AUTOSCROLL		 4 //!< Enable autoscroll flag. Check SetLayerAutoScroll() for further details.
#define TE_LAYERFLAG_UPDATE			 8 //!< Enable the updating flag. Check Update() for further details.
#define TE_LAYERFLAG_UPDATECELLS    16 //!< Enable the updating cells flag. Check Update() for further details.
#define TE_LAYERFLAG_RENDER			32 //!< Enable the rendering flag. Check Render() for further details. 
#define TE_LAYERFLAG_SMOOTHSCROLL	64 //!< Enable the smooth rendering flag. Check Render() for further details.
#define TE_LAYERFLAG_SHOWGRID	   128 //!< Show the base mapdata grid (for debugging purposes). Check SetLayerGridColor() for further details.
#define TE_LAYERFLAG_PAUSE		   256 //!< Pause the layer (current positions can not be modified). Update() and Render() methods should be called to keep the internal state.

#define TE_KEEP_VALUE           0x00FFFFFE //!< Used with properties members to do not change the previous value
#define TE_LAYERPOSITION_BEGIN  0x00FFFFFD //!< Set the layer position at the beginning
#define TE_LAYERPOSITION_MIDDLE 0x00FFFFFC //!< Set the layer position at the middle
#define TE_LAYERPOSITION_END    0x00FFFFFB //!< Set the layer position at the end
#define TE_LAYERPOSITION_CENTER 0x0A000000 //!< Added to a requested layer position to center on it

#define TE_RESCONTROL_USER				 0 //!< Used for layer maps and tilesets. The user must remove these resources after usage. Can be modified for layers but not for tilesets.
#define TE_RESCONTROL_AUTO				 1 //!< Used for layer maps and tilesets. The engine will automatically remove these resources. Can be modified for layers but not for tilesets.
#define TE_RESCONTROL_INTERNAL			 2 //!< Used for layer maps and tilesets. Importing or loading levels will set on this flag. The engine automatically manages the resources. Can not be modified.

#define TE_REGISTEREDMAPDATA_MAX		 3 //!< Maximum registered mapdata types. [0] is always cMSTE_MapBase.
#define TE_TILESET_DEFAULT_SOURCE		"TE-TS-DEFAULT-SOURCE" //!< Default source for loaded tilesets. It sets that at saving level time, the tileset will be saved in the same location as the level.
#define TE_LAYER_EMPTY					"TE-LAYER-EMTPY" //!< Default name for a layer which is allocated but empty

//-------------------------------------------------------------------------------
// MegaStorm TileEngine main class.
//-------------------------------------------------------------------------------
//! MegaStorm Tile-based Engine
/*! Main scroll engine interface.*/ 
class cMSTE
{
    friend class cMSTE_MapBase; // CellUpdate() and CellRender() can access to the engine instance
private:
    // OnLayerEnd global definition. 
    typedef int(*TE_OnLayerEnd)(int cellX,int cellY, int startX,int endX,int incX, int startY,int endY,int incY); 

public:
    // General stuff methods (core)
    cMSTE();
    ~cMSTE();	
    int  Init(int iNumLayers = 3, char *levelname = "");
    int  Quit();	
    int  AllocateLayers(int iN);	
    int  FreeLayer(int layer);
    int  FreeLevel();
    void InfoPrint(int layer = -1);
    void LogVerbosity(int);

    // Setup engine methods (core)
    int			SetViewport(SDL_Rect *rect);
    SDL_Rect	*GetViewport();
    int			SetRenderTarget(SDL_Surface *surf);
    SDL_Surface	*GetRenderTarget();
    void		SetName(char *levelname);
    char		*GetName();
    char		*GetModificationDate();
    char		*GetCreationDate();
    int			RegisterMapDataHandler(cMSTE_MapBase *(*createInstance)(int y, int x));
    int         AddCustomMapData(int layer, cMSTE_MapBase *map);
    int         DeleteCustomMapData(int layer);

    // Move, update, render and conversion methods (core)
    int MoveLayer(int layer, float x, float y);
    int Update(int layer = -1);	
    int Render(int layer = -1);	
    int Screen2LayerCell(int layer, int sx, int sy, int *cx, int *cy);
    int Screen2LayerAbsolute(int layer, int sx, int sy, int *px, int *py);

    // Layer properties methods (core)
    int             SetLayerFlags(int layer,Uint32 iFlags, char iState = 0);
    Uint32          GetLayerFlags(int layer);
    int             SetLayerResControl(int layer, char rescontrol);
    char            GetLayerResControl(int layer);	
    int             SetLayerAlpha(int layer, int alpha);
    int				GetLayerAlpha(int layer);	
    int             SetLayerParallaxRatio(int layer, float pratio);
    float           GetLayerParallaxRatio(int layer);
    int             SetLayerPosition(int layer, float xpos = TE_KEEP_VALUE, float ypos = TE_KEEP_VALUE);
    int             GetLayerPosition(int layer, float & xpos, float & ypos);
    int             SetLayerAutoScroll(int layer, float xspeed = TE_KEEP_VALUE, float yspeed = TE_KEEP_VALUE);
    int             GetLayerAutoScroll(int layer, float & xspeed, float & yspeed);
    int             SetLayerMapData(int layer, cMSTE_MapBase *map);
    cMSTE_MapBase  *GetLayerMapData(int layer);
    int             SetLayerOnUpdateEnd(int layer, TE_OnLayerEnd myfunc);
    TE_OnLayerEnd  *GetLayerOnUpdateEnd(int layer);
    int             SetLayerOnRenderEnd(int layer, TE_OnLayerEnd myfunc);
    TE_OnLayerEnd  *GetLayerOnRenderEnd(int layer);
    char		   *GetLayerName(int layer);
    int			    SetLayerName(int layer, char *name);
    int				SetLayerGridColor(int layer, Uint32 iColor);

    // Tileset handling (core)
    int   LoadTileset(CRM32Pro_CTile *tiles, int iID = -1);
    int   LoadTileset(char *sDPFfile, char *sTileName, int iID = -1);
    int   LoadTileset(char *sIMGfile, char *sTileName, int iTileWidth, int iTileHeight,int iID = -1);
    int   UnloadTileset(int iTN);
    CRM32Pro_CTile *GetTileset(int iTN);
    char *GetTilesetSource(int iTN);
    int   SetTilesetSource(int iTN, char *sSource);

    // IO Filters (IO_*)
    int SaveLevel(char *sLevelFilename, char *sDPF = NULL, char *sLevelname = NULL);
    int LoadLevel(char *sLevelFilename, char *sDPF = NULL, char *sLevelname = NULL);
    int ImportImage(int layer, SDL_Surface *surface, int tileSizeX, int tileSizeY, int accuracy = 100, int colorkey = -1);
    int ImportImage(int layer, char *fileimage, int tileSizeX, int tileSizeY, int accuracy = 100, int colorkey = -1);
    int ExportImage(int layer, SDL_Surface *&surface);
    int ExportImage(int layer, char *filename);
    int ExportTMX(char *sTMXFilename, char *sLevelname = NULL);
    int ImportTMX(char *sTMXFilename);

private:
    // Core methods (core)
    int            ValidateMapBase(int layer);                      // Validate the mapbase against available tilesets
    int			   CheckTile(unsigned int iTN);						// Check if the tile exists on any available tileset
    int            TryMoveX(int layer, float newpos);               // Try to move to new X position
    int            TryMoveY(int layer, float newpos);               // Try to move to new Y position
    cMSTE_MapBase *CreateMapDataInstance(char *type, int y, int x); // Create a mapdata instance of the given type

    //! Layer struct
    typedef struct st_layers
    {	
        Uint32	iStatus;			//!< See TE_LAYERFLAG_xxx defines
        char	bResControl;		//!< See TE_RESCONTROL_xxx defines
        char	bTilesetReady;		//!< Flag informing valid tilesets are loaded for our base mapdata
        float	fParallaxRatio;		//!< Parallax speed ratio 		
        float	fSpeedX,fSpeedY;	//!< X,Y autoscroll speed (pixels per second)
        int		iAlpha;				//!< The opacity of the layer as a value from 0 (transparent) to 255 (opaque). By default 255
        Uint32	iGridColor;			//!< Grid color for the base mapdata when TE_LAYERFLAG_SHOWGRID is set
        cMSTE_MapBase  *map;		//!< Pointer to a MapBase object
        TE_OnLayerEnd OnUpdateEnd;	//!< OnUpdateEnd event: your own update function to be called when Update() finishes to update this layer
        TE_OnLayerEnd OnRenderEnd;	//!< OnRenderEnd event: your own render function to be called when Render() finishes to render this layer
        char	szName[32];			//!< Layer name			
        // Constructors/destructor
        st_layers::st_layers();
        st_layers::~st_layers();
        void reset();
        // Internal vars		
        float fPosX, fPosY;            //!< Absolute current position (in pixels)
        float fLogicX,fLogicY;         //!< Absolute current logic position (in pixels)
        float fLogicX_pre,fLogicY_pre; //!< Absolute previous logic position (in pixels)
        float fGfxX, fGfxY;            //!< Absolute current graphics position (in pixels, for the SmoothScroll)
        int iOffsetX, iOffsetY;        //!< Used for rendering layers with tiles greater than the cell size
        Uint32 iLastUpdateTime;
        Uint32 iDisableNextSmooth;		
    };
    inline int ValidateMapX(int iX,st_layers *lP);   // Convert a raw X map position to a valid map position 
    inline int ValidateMapY(int iY,st_layers *lP);   // Convert a raw Y map position to a valid map position
    //int CalculateSmooth(st_layers::sPosition *, int, st_layers *); TO BE REMOVED

    //! Tilesets struct
    typedef struct st_tilesets
    {
        int iGID;					//!< Global ID of this tileset
        char	*szSource;          //!< String with the tileset source. It sets from where it comes (external image? or DPF?)
        CRM32Pro_CTile *tTileset;   //!< Pointer to the tileset. It contains the name, width, height, colorkey, alpha value and number of tiles.
        unsigned int iNumTiles;		//!< Number of tiles on the tileset
        char	 bResControl;		//!< See TE_RESCONTROL_xxx defines
        // Constructors/destructor
        st_tilesets::st_tilesets();
        st_tilesets::st_tilesets(int gID, char *Source, CRM32Pro_CTile *tile, char bRes);
        st_tilesets::~st_tilesets();
    };		

    // Private attributes
    char		bTileEngine;			//!< Main engine status flag -> 0: Needs init & render surface. 1: Needs render surface. 2: ready to go!
    SDL_Surface	*sRenderTarget;			//!< Destination surface pointer
    SDL_Rect	rViewport;				//!< Set the render viewport to be used by the TE
    vector<st_layers*>	vLayers;		//!< Vector of layers
    vector<st_tilesets*> vTilesets;		//!< Vector of tilesets
    int         iLogL,iLogN,iLogH;      
    char		szName[32];				//!< Level name
    char		szCreationDate[32];		//!< Level creation date
    char		szModificationDate[32];	//!< Level last modification date
    cMSTE_MapBase * (*RegisteredMapData[TE_REGISTEREDMAPDATA_MAX])(int y,int x); //!< List of pointers to registered mapdata handlers. [0] is always cMSTE_MapBase.
};
//-------------------------------------------------------------------------------
#endif

// Write a TMX property
extern int WriteTMXProperty(int idXML, char *name, char *value);
extern int WriteTMXProperty(int idXML, char *name, int value);
extern int WriteTMXProperty(int idXML, char *name, double value);

// Read a TMX property
extern int ReadTMXProperty(int idXML, char *name, char **value);
extern int ReadTMXProperty(int idXML, char *name, int *value);
extern int ReadTMXProperty(int idXML, char *name, double *value);

//! @}
//-------------------------------------------------------------------------------
// End of MegaStorm TileEngine header
//-------------------------------------------------------------------------------
